@@ -23,6 +23,9 @@
  */
 #include <config.h>
 
+#include <errno.h>
+#include <limits.h>
+#include <stdlib.h>
 #include <string.h>
 
 #include "theme-pixbuf.h"
@@ -34,7 +37,7 @@
 #define LOG(...)
 #endif
 
-static GCache *pixmap_cache = NULL;
+static GHashTable *pixbuf_hash = NULL;
 
 ThemePixbuf *
 theme_pixbuf_new (void)
@@ -54,22 +57,24 @@ theme_pixbuf_destroy (ThemePixbuf *theme_pb)
   else if (theme_pb->refcnt > 1)
       g_warning ("[%p] destroy: refcnt > 1", theme_pb);
 
-  theme_pixbuf_set_filename (theme_pb, NULL);
+  if (theme_pb->shared)
+    {
+      g_hash_table_remove (pixbuf_hash, theme_pb);
+      if (theme_pb->pixmap)
+	sapwood_pixmap_free (theme_pb->pixmap);
+    }
+  if (theme_pb->basename)
+    g_free (theme_pb->basename);
   g_free (theme_pb);
 }
 
 static ThemePixbuf *
-theme_pixbuf_copy (ThemePixbuf *theme_pb)
+theme_pixbuf_ref (ThemePixbuf *theme_pb)
 {
-  ThemePixbuf *copy;
-
-  copy = g_memdup(theme_pb, sizeof(*theme_pb));
-  copy->refcnt = 1;
-  copy->dirname  = theme_pb->dirname;
-  copy->basename = g_strdup (theme_pb->basename);
-  copy->pixmap = NULL;
-
-  return copy;
+  theme_pb->refcnt++;
+  LOG ("[%p] ref: refcnt = %d", theme_pb, theme_pb->refcnt);
+  g_assert(theme_pb->refcnt != 0);
+  return theme_pb;
 }
 
 void
@@ -98,55 +103,75 @@ theme_pixbuf_equal (gconstpointer v1, gconstpointer v2)
       !g_str_equal (a->basename, b->basename))
     return FALSE;
 
-  if (a->border_bottom != b->border_bottom ||
-      a->border_top != b->border_top ||
-      a->border_left != b->border_left ||
-      a->border_right != b->border_right)
+  return TRUE;
+}
+
+ThemePixbuf *
+theme_pixbuf_canonicalize (ThemePixbuf *theme_pb, gboolean *warn)
+{
+  ThemePixbuf *canon;
+
+  g_assert (theme_pb->pixmap == NULL);
+
+  *warn = FALSE;
+
+  if (!pixbuf_hash)
+    pixbuf_hash = g_hash_table_new (theme_pixbuf_hash, theme_pixbuf_equal);
+
+  canon = g_hash_table_lookup (pixbuf_hash, theme_pb);
+  if (!canon)
     {
-      g_warning ("file: %s", a->basename);
-      if (a->border_bottom != b->border_bottom)
-	g_warning ("border_bottom differs");
-      if (a->border_top != b->border_top)
-	g_warning ("border_top differs");
-      if (a->border_left != b->border_left)
-	g_warning ("border_left differs");
-      if (a->border_right != b->border_right)
-	g_warning ("border_right differs");
+      theme_pb->shared = 1;
+      g_hash_table_insert (pixbuf_hash, theme_pb, theme_pb);
+      canon = theme_pb;
+    }
+  else
+    {
+      if (theme_pb->border_bottom != canon->border_bottom ||
+          theme_pb->border_top    != canon->border_top ||
+          theme_pb->border_left   != canon->border_left ||
+          theme_pb->border_right  != canon->border_right ||
+          theme_pb->stretch       != canon->stretch)
+        *warn = TRUE;
+
+      theme_pixbuf_ref (canon);
+      theme_pixbuf_unref (theme_pb);
     }
 
-  return TRUE;
+  return canon;
 }
 
 void         
 theme_pixbuf_set_filename (ThemePixbuf *theme_pb,
 			   const char  *filename)
 {
-  if (theme_pb->pixmap)
-    {
-      g_cache_remove (pixmap_cache, theme_pb->pixmap);
-      theme_pb->pixmap = NULL;
-    }
+  g_assert (theme_pb->pixmap == NULL);
 
   if (theme_pb->basename)
     g_free (theme_pb->basename);
 
+  theme_pb->dirname  = NULL;
+  theme_pb->basename = NULL;
+
   if (filename)
     {
-      char *dirname;
-      char *basename;
+      char abspath[PATH_MAX + 1];
 
-      dirname  = g_path_get_dirname (filename);
-      basename = g_path_get_basename (filename);
+      if (realpath (filename, abspath))
+	{
+	  char *dirname;
+	  char *basename;
 
-      theme_pb->dirname  = g_quark_to_string (g_quark_from_string (dirname));
-      theme_pb->basename = basename;
+	  dirname  = g_path_get_dirname (abspath);
+	  basename = g_path_get_basename (abspath);
 
-      g_free (dirname);
-    }
-  else
-    {
-      theme_pb->dirname  = NULL;
-      theme_pb->basename = NULL;
+	  theme_pb->dirname  = g_quark_to_string (g_quark_from_string (dirname));
+	  theme_pb->basename = basename;
+
+	  g_free (dirname);
+	}
+      else
+	g_warning ("%s: %s", filename, g_strerror (errno));
     }
 }
 
@@ -174,47 +199,29 @@ theme_pixbuf_set_stretch (ThemePixbuf *theme_pb,
   theme_pb->stretch = stretch;
 }
 
-static SapwoodPixmap *
-pixmap_cache_value_new (ThemePixbuf *theme_pb)
-{
-  SapwoodPixmap *result;
-  char          *filename;
-  GError *err = NULL;
-
-  filename = g_build_filename (theme_pb->dirname, theme_pb->basename, NULL);
-  result = sapwood_pixmap_get_for_file (filename,
-						     theme_pb->border_left,
-						     theme_pb->border_right,
-						     theme_pb->border_top,
-						     theme_pb->border_bottom,
-						     &err);
-  if (!result)
-    {
-      g_warning ("sapwood-theme: Failed to load pixmap file %s: %s\n",
-		 filename, err->message);
-      g_error_free (err);
-    }
-
-  g_free (filename);
-
-  return result;
-}
-
 static SapwoodPixmap *
 theme_pixbuf_get_pixmap (ThemePixbuf *theme_pb)
 {
   if (!theme_pb->pixmap)
     {
-      if (!pixmap_cache)
-	pixmap_cache = g_cache_new ((GCacheNewFunc)pixmap_cache_value_new,
-				    (GCacheDestroyFunc)sapwood_pixmap_free,
-				    (GCacheDupFunc)theme_pixbuf_copy,
-				    (GCacheDestroyFunc)theme_pixbuf_unref,
-				    theme_pixbuf_hash, g_direct_hash, theme_pixbuf_equal);
+      char   *filename;
+      GError *err = NULL;
 
-      theme_pb->pixmap = g_cache_insert (pixmap_cache, theme_pb);
+      filename = g_build_filename (theme_pb->dirname, theme_pb->basename, NULL);
+      theme_pb->pixmap = sapwood_pixmap_get_for_file (filename,
+						      theme_pb->border_left,
+						      theme_pb->border_right,
+						      theme_pb->border_top,
+						      theme_pb->border_bottom,
+						      &err);
       if (!theme_pb->pixmap)
-	g_cache_remove (pixmap_cache, NULL);
+	{
+	  g_warning ("sapwood-theme: Failed to load pixmap file %s: %s\n",
+		     filename, err->message);
+	  g_error_free (err);
+	}
+
+      g_free (filename);
     }
   return theme_pb->pixmap;
 }
@@ -267,12 +274,12 @@ theme_pixbuf_render (ThemePixbuf  *theme_pb,
 
   pixmap = theme_pixbuf_get_pixmap (theme_pb);
 
-  /* if we do scaling we want to draw at least the whole pixmap */
-  draw_width  = MAX(width, pixbuf_width);
-  draw_height = MAX(height, pixbuf_height);
-
   if (theme_pb->stretch)
     {
+      /* if we do scaling we want to draw at least the whole pixmap */
+      draw_width  = MAX(width, pixbuf_width);
+      draw_height = MAX(height, pixbuf_height);
+
       dest_x[0] = x;
       dest_x[1] = x + theme_pb->border_left;
       dest_x[2] = x + draw_width - theme_pb->border_right;
@@ -329,14 +336,14 @@ theme_pixbuf_render (ThemePixbuf  *theme_pb,
 	       */
 	      if (clip_rect->width < mask_width)
 		{
-		  LOG("width: %d -> %d", mask_width, clip_rect->width);
-		  mask_x = clip_rect->x;
+		  LOG("width: %d -> %d, x: %d - %d", mask_width, clip_rect->width, x, clip_rect->x);
+		  mask_x = x - clip_rect->x;
 		  mask_width = clip_rect->width;
 		}
 	      if (clip_rect->height < height)
 		{
-		  LOG("height: %d -> %d", mask_height, clip_rect->height);
-		  mask_y = clip_rect->y;
+		  LOG("height: %d -> %d, y: %d - %d", mask_height, clip_rect->height, y, clip_rect->y);
+		  mask_y = y - clip_rect->y;
 		  mask_height = clip_rect->height;
 		}
 	    }
@@ -377,8 +384,12 @@ theme_pixbuf_render (ThemePixbuf  *theme_pb,
     }
   else if (center)
     {
-      x += (draw_width - pixbuf_width) / 2;
-      y += (draw_height - pixbuf_height) / 2;
+      /* when centering don't expand beyond pixbuf size */
+      draw_width = MIN(width, pixbuf_width);
+      draw_height = MIN(height, pixbuf_height);
+
+      x += (width - draw_width) / 2;
+      y += (height - draw_height) / 2;
 
       sapwood_pixmap_get_pixmap (pixmap, 1, 1,
                                  &rect[0].pixmap, &rect[0].pixmask);
@@ -388,7 +399,7 @@ theme_pixbuf_render (ThemePixbuf  *theme_pb,
       rect[0].dest.height = pixbuf_height;
 
       sapwood_pixmap_render_rects (pixmap,
-                                   window, x, y, width, height,
+                                   window, x, y, draw_width, draw_height,
                                    mask, x, y, FALSE,
                                    clip_rect, 1, rect);
     }
